"""Model references"""

import abc
import contextlib
import inspect
import operator
import typing
import warnings

import pydantic
import typing_extensions

KeyType = typing.TypeVar("KeyType")
UrlType = typing.TypeVar("UrlType")


def _extract_type(meth: typing.Callable) -> typing.Type:
    return_annotation = inspect.signature(meth).return_annotation
    assert (
        return_annotation != inspect.Signature.empty
    ), f"return annotation of {meth!r} unexpectedly empty"
    return return_annotation


class Referrable(typing_extensions.Protocol[KeyType, UrlType]):
    """Protocol that needs to be implemented by a target of :class:`Href`

    The class can either be used as a protocol (see `PEP 544
    <https://www.python.org/dev/peps/pep-0544/>`_).

    * When used as protocol in type annotations, :class:`Referrable` is
      parametrized by key and URL types, respectively. For example
      ``Referrable[int, str]`` annotates a referrable type having ``int`` as key
      and ``str`` as URL. ``Referrable[UUID, AnyHttpUrl]`` annotates a
      referrable type having ``UUID`` key and ``AnyHttpUrl`` as URL type.

    * When used as abstract base class, the subclass needs to implement at least
      :func:`key_to_url()` and :func:`url_to_key()` to specify the conversions
      between the key and URL representations. The return types of the functions
      should be annotated to make them available for parsing and serialization
      at runtime. Moreover, the default implementation assumes that the has
      ``id`` property used as the key of the referrable. Here is an example:

      .. code-block:: python

         class Book(Referrable):
             id: int

             @classmethod
             def key_to_url(key: int) -> str:
                 return f"/books/{key}"

             @classmethod
             def url_to_key(url: str) -> int:
                 return url.split("/")[-1]

    """

    def get_key(self) -> KeyType:
        """Return the key of the model

        The default implementation returns the ``id`` property of the object.
        """
        return getattr(self, "id")

    @classmethod
    def href_types(cls) -> typing.Tuple[typing.Type[KeyType], typing.Type[UrlType]]:
        """Return a tuple containing the key and url types, respectively

        The default implementation returns the return type annotations of
        :func:`url_to_key()` and :func:`key_to_url()`, respectively.
        """
        return _extract_type(cls.url_to_key), _extract_type(cls.key_to_url)

    @classmethod
    @abc.abstractmethod
    def key_to_url(cls, key: KeyType) -> UrlType:
        """Convert key to url"""
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def url_to_key(cls, url: UrlType) -> KeyType:
        """Convert url to key"""
        raise NotImplementedError()


ReferrableType = typing.TypeVar("ReferrableType", bound=Referrable)


class Href(typing.Generic[ReferrableType]):
    """Hypertext reference to another model

    The class is generic and can be annotated by a type implementing the
    :class:`Referrable` protocol. If ``Book`` is assumed to be a type
    implementing :class:`Referrable`, then ``Href[Book]`` represents a hyperlink
    to a book. This mechanism primarily exists for the benefit of pydantic, and
    allows the validation to know what kind of reference it is working with (see
    :ref:`quickstart`).

    """

    __slots__ = ["_key", "_url"]

    def __init__(self, key, url):
        """
        Arguments:
          key: the key used by the application to identify the model internally
          url: the URL identifying the model externally (e.g. via REST API)
        """
        self._key = key
        self._url = url

    def __repr__(self):
        return f"Href(key={self._key!r}, url={self._url!r})"

    @property
    def key(self):
        """The key of the referred object"""
        return self._key

    @property
    def url(self):
        """The URL of the referred object"""
        return self._url

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value, field: pydantic.fields.ModelField):
        """Parse ``value`` into hyperlink

        This method mainly exists for integration to pydantic. A user rarely
        needs to call it directly.

        Arguments:
          value:

            The parsed object. It can be either:

            * Another :class:`Href` instance
            * An instance of the referred model
            * A value of the key type (interpreted as key identifying the referred object)
            * A url string (interpreted as URL to the referred objet)

        Returns:
          A :class:`Href` object referring the model identified by the ``value``
          argument.

        Raises:
          TypeError: If the :class:`Href` model isn't properly annotated, or if
            ``value`` doesn't conform to any of the recognized types
          Exception: In addition passes any exceptions happening during conversions

        """
        if not field.sub_fields:
            raise TypeError("Expected sub field")
        model_type: typing.Type[ReferrableType] = field.sub_fields[0].type_
        if isinstance(value, cls):
            return value
        if isinstance(value, model_type):
            key = value.get_key()
            return cls._from_key(key, model_type)
        key_type, url_type = model_type.href_types()
        with contextlib.suppress(pydantic.ValidationError):
            key = pydantic.parse_obj_as(key_type, value)
            return cls._from_key(key, model_type)
        with contextlib.suppress(pydantic.ValidationError):
            url = pydantic.parse_obj_as(url_type, value)
            return cls._from_url(url, model_type)
        raise TypeError(f"Could not convert {value!r} to either key or url")

    @classmethod
    def _from_key(cls, key: KeyType, model_type: typing.Type[ReferrableType]):
        return cls(key=key, url=model_type.key_to_url(key))

    @classmethod
    def _from_url(cls, url: UrlType, model_type: typing.Type[ReferrableType]):
        return cls(key=model_type.url_to_key(url), url=url)


try:
    from pydantic.json import ENCODERS_BY_TYPE
except ImportError:
    warnings.warn(
        "Failed to add Href encoder. This may affect serializing Href instances to json.",
        ImportWarning,
    )
else:
    ENCODERS_BY_TYPE[Href] = operator.attrgetter("url")



from typing import List, Dict

from fastapi import FastAPI, HTTPException, Request, Response
from fastapi.middleware import Middleware
from pydantic import BaseModel

from hrefs import Href
from hrefs.starlette import ReferrableModel, HrefMiddleware


class Book(ReferrableModel):
    id: int
    title: str

    class Config:
        details_view = "get_book"


class Library(BaseModel):
    id: int
    books: List[Href[Book]]


books: Dict[int, Book] = {}
libraries: Dict[int, Library] = {}

app = FastAPI(middleware=[Middleware(HrefMiddleware)])


@app.get("/libraries/{id}", response_model=Library)
def get_library(id: int):
    library = libraries.get(id)
    if not library:
        raise HTTPException(status_code=404, detail="Library not found")
    return library


@app.post("/libraries")
def post_library(library: Library, request: Request):
    if any(book.get_key() not in books for book in library.books):
        raise HTTPException(
            status_code=400, detail="Trying to add nonexisting book to library"
        )
    libraries[library.id] = library
    return Response(
        status_code=201,
        headers={"Location": request.url_for("get_library", id=library.id)},
    )


@app.get("/books/{id}", response_model=Book)
def get_book(id: int):
    book = books.get(id)
    if not book:
        raise HTTPException(status_code=404, detail="Book not found")
    return book


@app.post("/books", status_code=204)
def post_book(book: Book, request: Request):
    books[book.id] = book
    return Response(
        status_code=201, headers={"Location": request.url_for("get_book", id=book.id)}
    )



"""Test utilities"""

import typing

from hrefs import Href, Referrable

from hypothesis import strategies as st


def hrefs(target: typing.Type[Referrable], key_strategy):
    key_type, _ = target.href_types()
    return key_strategy.map(lambda key: Href(key=key, url=target.key_to_url(key)))



"""Base models for referring and referrable types"""

import typing

import pydantic

from .href import Href, Referrable


if typing.TYPE_CHECKING:
    # `mypy` doesn't accept the true metaclass. should be investigated later.
    class BaseReferrableModel(pydantic.BaseModel, Referrable):
        """Dummy"""


else:
    # pylint: disable-all
    class _ReferrableModelMeta(pydantic.BaseModel.__class__, Referrable.__class__):
        pass

    class BaseReferrableModel(
        pydantic.BaseModel, Referrable, metaclass=_ReferrableModelMeta
    ):
        """Referrable model with pydantic integration

        A subclass of both :class:`pydantic.BaseModel` and
        :class:`hrefs.Referrable`.  It should be used as the base class of any
        pydantic model that will be used as target of :class:`hrefs.Href`.

        When using referrable models with FastAPI or Starlette in particular,
        :class:`hrefs.starlette.ReferrableModel` is more natural base.
        """



import json
import typing

from hypothesis import given, strategies as st
import pydantic
import pytest

from hrefs import Href, BaseReferrableModel
from util import hrefs


class Pet(BaseReferrableModel):
    id: int

    @staticmethod
    def key_to_url(key: int) -> str:
        return f"/pets/{key}"

    @staticmethod
    def url_to_key(url: str) -> int:
        return int(url.split("/")[-1])


class Owner(pydantic.BaseModel):
    pets: typing.List[Href[Pet]]


pet_hrefs = hrefs(Pet, st.integers())


@given(pet_hrefs)
def test_parse_href(href):
    assert pydantic.parse_obj_as(Href[Pet], href) is href


@given(st.builds(Pet))
def test_parse_referrable_model(pet):
    href = pydantic.parse_obj_as(Href[Pet], pet)
    assert href.key == pet.id
    assert href.url == Pet.key_to_url(pet.id)


@given(st.integers())
def test_parse_key_to_href(key):
    href = pydantic.parse_obj_as(Href[Pet], key)
    assert href.key == key
    assert href.url == Pet.key_to_url(key)


@given(st.from_regex(r"\A/pets/\d+\Z"))
def test_parse_url_to_key(url):
    href = pydantic.parse_obj_as(Href[Pet], url)
    assert href.key == Pet.url_to_key(url)
    assert href.url == url


def test_parse_href_with_unparseable_key_fails():
    with pytest.raises(pydantic.ValidationError):
        pydantic.parse_obj_as(Href, object())


def test_parse_href_without_parameter_fails():
    with pytest.raises(pydantic.ValidationError):
        pydantic.parse_obj_as(Href, 123)


@given(st.builds(Owner, pets=st.lists(pet_hrefs)))
def test_json_encode(owner):
    owner_json = json.loads(owner.json())
    assert owner_json["pets"] == [pet.url for pet in owner.pets]



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['hrefs']

package_data = \
{'': ['*'],
 'hrefs': ['.pytest_cache/*', '.pytest_cache/v/*', '.pytest_cache/v/cache/*']}

install_requires = \
['pydantic>=1.7', 'starlette']

extras_require = \
{'docs': ['sphinx', 'sphinx-rtd-theme'],
 'test': ['pytest>=6.2',
          'hypothesis>=6.24',
          'fastapi',
          'black',
          'pylint',
          'mypy']}

setup(name='hrefs',
      version='0.1',
      description='Reference support for pydantic models',
      author=None,
      author_email='Jaakko Moisio <jaakko@moisio.fi>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.7',
     )



import contextvars
import json
import random
import typing
import uuid

import fastapi
import fastapi.middleware
import fastapi.testclient
from hypothesis import given, strategies as st
import pydantic

from hrefs import Href
from hrefs.starlette import ReferrableModel, HrefMiddleware


class Comment(ReferrableModel):
    id: uuid.UUID

    class Config:
        details_view = "get_comment"


class Article(pydantic.BaseModel):
    id: uuid.UUID
    comments: typing.List[Href[Comment]]


article_var: contextvars.ContextVar[uuid.UUID] = contextvars.ContextVar("article_var")
comments_var: contextvars.ContextVar[typing.List[uuid.UUID]] = contextvars.ContextVar(
    "comments_var"
)
save_article_var: contextvars.ContextVar[
    typing.Callable[[Article], None]
] = contextvars.ContextVar("save_article_var")


app = fastapi.FastAPI(middleware=[fastapi.middleware.Middleware(HrefMiddleware)])


@app.get("/articles/{id}", response_model=Article)
async def get_article(id: uuid.UUID):
    assert id == article_var.get()
    return Article(
        id=id,
        comments=comments_var.get(),
    )


@app.post("/articles")
async def post_article(article: Article):
    save_article = save_article_var.get()
    save_article(article)


@app.get("/comments/{id}", response_model=Comment)
async def get_comment(id: uuid.UUID):
    assert id in comments_var.get()
    return Comment(id=id)


client = fastapi.testclient.TestClient(app)


@given(st.uuids(), st.lists(st.uuids(), min_size=1))
def test_parse_key_to_href(article_id, comment_ids):
    article_var.set(article_id)
    comments_var.set(comment_ids)
    response = client.get(f"/articles/{article_id}")
    assert response.status_code == fastapi.status.HTTP_200_OK, response.text
    article = response.json()
    comment_href, comment_id = random.choice(
        list(zip(article["comments"], comment_ids))
    )
    comment = client.get(comment_href).json()
    assert uuid.UUID(comment["id"]) == comment_id


@given(st.uuids(), st.lists(st.uuids(), min_size=1))
def test_parse_url_to_href(article_id, comment_ids):
    def assert_article(article: Article):
        assert article.id == article_id
        assert [comment_href.key for comment_href in article.comments] == comment_ids

    save_article_var.set(assert_article)
    response = client.post(
        "/articles",
        data=json.dumps(
            dict(
                id=str(article_id),
                comments=[f"http://testclient/comments/{id}" for id in comment_ids],
            )
        ),
    )
    assert response.status_code == fastapi.status.HTTP_200_OK, response.text


@given(st.uuids(), st.lists(st.uuids(), min_size=1))
def test_parse_invalid_url_fails(article_id, comment_ids):
    response = client.post(
        "/articles",
        data=json.dumps(
            dict(
                id=str(article_id),
                comments=[
                    f"http://testclient/not/a/real/route/{id}" for id in comment_ids
                ],
            )
        ),
    )
    assert response.status_code == fastapi.status.HTTP_422_UNPROCESSABLE_ENTITY



"""Starlette integration"""

import contextvars

import pydantic
import starlette.requests
import starlette.middleware.base

from .model import BaseReferrableModel


_request_var: contextvars.ContextVar[
    starlette.requests.Request
] = contextvars.ContextVar("_request_var")


class HrefMiddleware(starlette.middleware.base.BaseHTTPMiddleware):
    """Middleware for resolving hrefs

    This middleware needs to be added to the middleware stack of the Starlette
    app using the :class:`ReferrableModel`.
    """

    @staticmethod
    async def dispatch(request: starlette.requests.Request, call_next):
        _request_var.set(request)
        return await call_next(request)


class ReferrableModel(BaseReferrableModel):
    """Referrable model with Starlette integration

    This class implements the :class:`hrefs.BaseReferrableModel` with the
    following features:

    * Its key type is inferred from the ``id`` field of the subclass

    * Its URL type is always :class:`pydantic.AnyHttpUrl`

    * When converting between key and URL, it relies on the Starlette request
      objects. Thus a :class:`ReferrableModel` is only usable within request
      handlers or middleware above the :class:`HrefMiddleware` in the middleware
      stack of the application.

    Here is a minimal example of a model having key type ``int`` (inferred from
    the ``id`` property), and using route called ``"my_view"`` to convert
    to/from URLs.

    .. code-block:: python

        class MyModel(ReferrableModel):
            id: int

            class Config:
                details_view = "my_view"

    For a more complete example of using :mod:`hrefs` library with Starlette,
    please refer to :ref:`quickstart`.

    """

    @classmethod
    def href_types(cls):
        return cls._get_key_type(), pydantic.AnyHttpUrl

    @classmethod
    def key_to_url(cls, key):
        request = _request_var.get()
        details_view = cls._get_details_view()
        return pydantic.parse_obj_as(
            pydantic.AnyHttpUrl, request.url_for(details_view, id=key)
        )

    @classmethod
    def url_to_key(cls, url: pydantic.AnyHttpUrl):
        request = _request_var.get()
        details_view = cls._get_details_view()
        for route in request.app.routes:
            if route.name == details_view:
                _, scope = route.matches(
                    {"type": "http", "method": "GET", "path": url.path}
                )
                if scope:
                    return pydantic.parse_obj_as(
                        cls._get_key_type(), scope["path_params"]["id"]
                    )
        raise ValueError(f"Could not resolve {url} into key")

    @classmethod
    def _get_key_type(cls):
        return cls.__fields__["id"].type_

    @classmethod
    def _get_details_view(cls):
        return getattr(cls.__config__, "details_view")
